
#include "Context.h"
#include "Item.h"

Context::Context() {   // root context
  m_bB = m_bEM = m_bI = m_bS = m_bTT = false;
  m_nU = 0;

  m_nSize = -1;
  m_eColor = ITEM_NUM_ENUMS;
}

bool Context::operator==(Context const &rhs) const {
  return ( m_bB == rhs.m_bB &&
       m_bEM == rhs.m_bEM &&
       m_bI == rhs.m_bI &&
       m_bS == rhs.m_bS &&
       m_bTT == rhs.m_bTT &&
       m_nU == rhs.m_nU &&
       m_nSize == rhs.m_nSize &&
       m_eColor == rhs.m_eColor);
}

Context Context::TransformContext (Context const *pSource, DocItemHandle pItem) {
  if (!pItem->IsTag()) {
    return *pSource;
  }
  Context result(*pSource);
  DocItem *pItem2 = pItem.Pointer();
  if (pItem->m_bStart) {
    switch (pItem->m_ItemType) {  // start tag
      case ITEM_TAG_B:
        result.m_bB = true;
        break;
      case ITEM_TAG_EM:
        result.m_bEM = !result.m_bEM;
        if (result.m_bS) {
          result.m_bEM = true;    // S hides EM
        }
        break;
      case ITEM_TAG_I:
        result.m_bI = true;
        break;
      case ITEM_TAG_PL: 
        result.m_nU = 0;
        result.m_bB = result.m_bEM = result.m_bI = result.m_bS = result.m_bTT = false;
        break;
      case ITEM_TAG_S:
        result.m_bS = true;
        result.m_bEM = true;      // S hides EM
        break;
      case ITEM_TAG_TT:
        result.m_bTT = true;
        break;
      case ITEM_TAG_U:
        result.m_nU++;
        if (result.m_nU > 3) {
          result.m_nU = 3;
        }
        break;
      case ITEM_TAG_0:
      case ITEM_TAG_1:
      case ITEM_TAG_2:
      case ITEM_TAG_3:
      case ITEM_TAG_4:
      case ITEM_TAG_5:
      case ITEM_TAG_6:
      case ITEM_TAG_7:
      case ITEM_TAG_8:
      case ITEM_TAG_9:
        result.m_nSize = ((int) (pItem->m_ItemType)) - ((int) ITEM_TAG_0);
        break;
      case ITEM_TAG_r:
      case ITEM_TAG_g:
      case ITEM_TAG_b:
      case ITEM_TAG_c:
      case ITEM_TAG_m:
      case ITEM_TAG_y:
      case ITEM_TAG_k:
      case ITEM_TAG_w:
        result.m_eColor = pItem->m_ItemType;
        break;
      default:
        assert(false);
    }
  } else {          // end tag
    return pItem->GetOtherItem()->m_Context;
  }
  return result;    
}

bool Context::WhitespaceCompressible() const {
  if (m_nU > 0) {
    return false;
  }
  if (m_bTT) {
    return false;
  }
  return true;
}

void Context::print (FILE *file) const {
  fprintf(file, "B: %d EM: %d I: %d S: %d TT: %d U: %d size: %d color: %d\n",
    m_bB, m_bEM, m_bI, m_bS, m_bTT, m_nU, m_nSize, m_eColor);
}

bool Context::CompareTextOutput (ConstListNode begin1, ConstListNode end1, ConstListNode begin2, ConstListNode end2) {
  bool bEmpty1, bEmpty2;
  bEmpty1 = (begin1 == end1);
  bEmpty2 = (begin2 == end2);
  if (bEmpty1 && bEmpty2) {
    return true;
  }
  if (bEmpty1 || bEmpty2) {
    return false;
  }

  // make copies of context
  Context c1((*begin1)->m_Context);
  Context c2((*begin2)->m_Context);  

  if (TextContextEquivalent(c1, c2)) {
    string s1, s2;
    Document::CollectString(begin1, end1, true, s1);
    Document::CollectString(begin2, end2, true, s2);
    return (s1 == s2);
  }
  return false;
}

bool Context::CompareWhitespaceOutput (ConstListNode begin1, ConstListNode end1, ConstListNode begin2, ConstListNode end2) {
  bool bEmpty1, bEmpty2;
  bEmpty1 = (begin1 == end1);
  bEmpty2 = (begin2 == end2);
  if (bEmpty1 && bEmpty2) {
    return true;
  }
  if (bEmpty1 || bEmpty2) {
    return false;
  }

  DocItem *pItem1 = (*begin1).Pointer();
  DocItem *pItem2 = (*begin2).Pointer();

  // make copies of context
  Context c1(WhitespaceContext((*begin1)->m_Context));
  Context c2(WhitespaceContext((*begin2)->m_Context));  

  // ignore TT for a moment
  bool bTT1 = c1.m_bTT;
  c1.m_bTT = false;
  bool bTT2 = c2.m_bTT;
  c2.m_bTT = false;
  if (c1 != c2) {     // contexts are still different
    return false;
  }
  c1.m_bTT = bTT1;    // restore TT
  c2.m_bTT = bTT2;

  string s1, s2;
  if (c1.WhitespaceCompressible()) {
    s1 = " ";
  } else {
    Document::CollectString(begin1, end1, true, s1);
  }
  if (c2.WhitespaceCompressible()) {
    s2 = " ";
  } else {
    Document::CollectString(begin2, end2, true, s2);
  }
  return (s1 == s2);
}

bool Context::TextContextEquivalent (Context const &c1, Context const &c2) {
  return (c1 == c2);
}

bool Context::WhitespaceContextEquivalent (Context const &c1, Context const &c2) {
  return (WhitespaceContext(c1) == WhitespaceContext(c2));
}

Context Context::WhitespaceContext (Context const &c) {
  Context result(c);
  result.m_bS = result.m_bEM = result.m_bI = result.m_bB = false;     // unset s, em i, b 
  if (result.m_nU == 0) {
    result.m_eColor = ITEM_TAG_w;
  }
  return result;
}

