
#include <stdio.h>
#include <string>
#include <map>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "Parser.h"
#include "Globals.h"

#define TAG_START '<'
#define TAG_END '>'

Parser::Parser(ParserCallback *pTextCallback, ParserCallback *pWhitespaceCallback, ParserCallback *pDefaultTagCallback) {
  m_pTextCallback = pTextCallback;
  m_pWhitespaceCallback = pWhitespaceCallback;
  m_pDefaultTagCallback = pDefaultTagCallback;
  m_nLineno = 0;
  m_nBufSize = 1024;
  m_pcBuf = new char[1024];
}

void Parser::ResizeBuf(const char *pcSource, int nSize) {
  if (nSize + 1 > m_nBufSize) {
    delete [] m_pcBuf;
    m_pcBuf = new char[nSize + 1];
    m_nBufSize = nSize + 1;
  }
  memcpy(m_pcBuf, pcSource, nSize);
  m_pcBuf[nSize] = '\0';
}

void Parser::registerCallback(const char *pcTag, ParserCallback *pTagCallback, void *pArg) {
  m_CallbackMap[pcTag] = new CallbackInfo(pTagCallback, pArg);
}

void Parser::parse(const char *pSource, int nSize) {
  const char *c = pSource;
  const char *pEnd = pSource + nSize;
  const char *pcStart = pSource;
  int nCharPos = 0;
  int state = 0;  // 0 - start
                  // 1 - in text
                  // 2 - in whitespace
                  // 3 - in tag
  while (c != pEnd) {
    switch (state) {
      case 0:
        pcStart = c;
        if (*c == TAG_START) {        // start -> tag start
          state = 3;
        } else if (isspace(*c)) {     // start -> whitespace
          state = 2;
        } else if (*c == TAG_END) {   // start -> tag end
          MyError("unmatched >", m_nLineno, nCharPos);
        } else {                      // start -> text
          state = 1;
        }
        break;
      case 1:
        if (*c == TAG_START) {        // text -> tag start
          HandleText(pcStart, c - pcStart, m_nLineno);
          state = 3;
          pcStart = c;
        } else if (isspace(*c)) {     // text -> whitespace
          HandleText(pcStart, c - pcStart, m_nLineno);
          state = 2;
          pcStart = c;
        } else if (*c == TAG_END) {   // text -> tag end
          MyError("unmatched >", m_nLineno, nCharPos);
        } else {                      // text -> text
        }
        break;
      case 2:
        if (*c == TAG_START) {        // whitespace -> tag start
          HandleWhitespace(pcStart, c - pcStart, m_nLineno);
          state = 3;
          pcStart = c;
        } else if (isspace(*c)) {     // whitespace -> whitespace
        } else if (*c == TAG_END) {   // whitespace -> tag end
          MyError("unmatched >", m_nLineno, nCharPos);
        } else {                      // whitespace -> text
          HandleWhitespace(pcStart, c - pcStart, m_nLineno);
          state = 1;
          pcStart = c;
        }
        break;        
      case 3:
        if (*c == TAG_START) {        // tag -> tag start
          MyError("< in tag", m_nLineno, nCharPos);
        } else if (isspace(*c)) {     // tag -> whitespace
          MyError("unmatched <", m_nLineno, nCharPos);
        } else if (*c == TAG_END) {   // tag -> tag end
          HandleTag(pcStart, c - pcStart + 1, m_nLineno);
          state = 0;
          pcStart = c;
        } else {                      // tag -> text
        }
        break;
      default:
        assert(false);
    }
    nCharPos++;
    if (*c == '\n') {
      m_nLineno++;
      nCharPos = 0;
    }
    c++;
  }
  switch (state) {   // eof
    case 0:
      break;
    case 1:
      HandleText(pcStart, c - pcStart, m_nLineno);
      break;
    case 2:
      HandleWhitespace(pcStart, c - pcStart, m_nLineno);
      break;
    case 3:
      MyError("unexpected end of file", m_nLineno, nCharPos);
      break;
    default:
      assert(false);
  }
  return;
} 

void Parser::HandleText (const char *pcSource, int nSize, int nLineno) {
  if (m_pTextCallback != NULL) {
    ResizeBuf(pcSource, nSize);
    (*m_pTextCallback)(NULL, m_pcBuf, nSize, nLineno, false);
  }
}

void Parser::HandleWhitespace (const char *pcSource, int nSize, int nLineno) {
  if (m_pWhitespaceCallback != NULL) {
    ResizeBuf(pcSource, nSize);
    (*m_pWhitespaceCallback)(NULL, m_pcBuf, nSize, nLineno, false);
  }
}

void Parser::HandleTag (const char *pcSource, int nSize, int nLineno) {
  assert(nSize > 2);
  ResizeBuf(pcSource, nSize);
  m_pcBuf[nSize-1] = '\0';  // remove '>'
  CallbackMap::iterator item;
  if (m_pcBuf[1] != '/') {  // start tag
    item = m_CallbackMap.find(m_pcBuf + 1);
    if (item != m_CallbackMap.end()) {
      ParserCallback *pCallback = (*item).second->m_pCallback;
      void *pArg = (*item).second->m_pArg;
      (*pCallback)(pArg, m_pcBuf + 1, nSize, nLineno, true);
    } else {
      if (m_pDefaultTagCallback != NULL) {
        (*m_pDefaultTagCallback)(NULL, m_pcBuf + 1, nSize, nLineno, true);
      }
    }
  } else {
    item = m_CallbackMap.find(m_pcBuf + 2);
    if (item != m_CallbackMap.end()) {
      ParserCallback *pCallback = (*item).second->m_pCallback;
      void *pArg = (*item).second->m_pArg;
      (*pCallback)(pArg, m_pcBuf + 2, nSize, nLineno, false);
    } else {
      if (m_pDefaultTagCallback != NULL) {
        (*m_pDefaultTagCallback)(NULL, m_pcBuf + 2, nSize, nLineno, false);
      }
    }
  }
}

