/*
 *  RAY++ - Object-oriented ray tracing library
 *  Copyright (C) 1998 Martin Reinecke, Andreas Kahler
 *  and Neeharika Adabala
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  See the README file for more information.
 */

#ifndef RAYPP_HANDLE_H
#define RAYPP_HANDLE_H

#include <sys/types.h>
#include <assert.h>
#include <stdio.h>

#ifndef ulong_t
  typedef unsigned long ulong_t;
#endif

//#include "config/config.h"
//#include "kernel/error.h"

//RAYPP_NAMESPACE_BEGIN (RAYPP)

/**
  \class HANDLE kernel/handle.h kernel/handle.h
  Pointer with reference counting.
  This class implements a 'smart pointer' with reference count and
  automatic deallocation after deletion of the last reference.
*/
inline void _handle_error(const char *pcMsg) {
  printf ("handle reference error: %s\n", pcMsg);
  assert(false);
}
template <class T> class HANDLE
  {
  private:
    class HANDLE_DATA
      {
      private:
// prevent copying
        HANDLE_DATA (const HANDLE_DATA &);
        HANDLE_DATA &operator=(const HANDLE_DATA &);

      public:
        T *data;
        ulong_t counter;

        HANDLE_DATA (T *val)
          : data (val), counter (1) {}
        ~HANDLE_DATA ()
          { delete data; }
      };

    HANDLE_DATA *HandlePtr;

  public:
    /*!
       Default constructor. Handle points to 0.
      */
    HANDLE () 
      : HandlePtr (0) {}
    /*!
       Handle points to the same object as orig.
      */
    HANDLE (const HANDLE &orig)
      : HandlePtr (orig.HandlePtr)
      { if (HandlePtr) ++(HandlePtr->counter); }
    /*!
       Handle is initialized with contents; reference count is set to 1.
       \warning contents must not point to an automatic variable!
       After calling this constructor, this pointer should not be used again in
       any form! <br>
       Recommended usage of this constructor:
       \code HANDLE<A> foo (new A); \endcode
       Using it this way, one can make no mistakes.
      */
    HANDLE (T *contents)
      {
      if (contents)
        {
        HandlePtr = new HANDLE_DATA (contents);
        contents = 0;
        }
      else HandlePtr = 0;
      }
    /*!
       Destructor. If this is the last reference to the object pointed to,
       it is deleted.
      */
    ~HANDLE ()
      {
      if (HandlePtr)
        if (--(HandlePtr->counter) == 0) delete HandlePtr;
      }

    /*!
       The handle is unbound from the object it formerly pointed to
       (this object is eventually destroyed) and bound to the object
       pointed to by orig.
      */
    HANDLE &operator= (const HANDLE &orig)
      {
      if (orig.HandlePtr) ++(orig.HandlePtr->counter);
      if (HandlePtr)
        if (--(HandlePtr->counter) == 0) delete HandlePtr;
      HandlePtr = orig.HandlePtr;
      return *this;
      }    
    /*!
       The handle is unbound from the object it formerly pointed to
       (this object is eventually destroyed) and bound to the object
       pointed to by contents. <br>
       \warning contents must not point to an automatic variable!
       After calling this function, this pointer should not be used again in
       any form! <br>
       Recommended usage of this function:
       \code HANDLE<A> foo; foo = new A; \endcode
       using it this way, one can make no mistakes.
      */
    HANDLE &operator= (T *contents)
      {
      if (HandlePtr)
        if (--(HandlePtr->counter) == 0) delete HandlePtr;
      if (contents) HandlePtr = new HANDLE_DATA (contents);
      else HandlePtr = 0;
      return *this;
      }
    /*!
       The handle is unbound from the object it formerly pointed to
       (this object is eventually destroyed) and set to 0.
      */
    void Clear ()
      {
      if (HandlePtr)
        {
        if (--(HandlePtr->counter) == 0) delete HandlePtr;
        HandlePtr = 0;
        }
      }

    /*!
       Returns a pointer to the handle's object.
       If the handle points to 0, program execution is stopped.
      */
    T *operator-> ()
      {
      if (HandlePtr) return HandlePtr->data;
      _handle_error ("HANDLE: invalid operation");
// kludge: Some compilers want a return value
      return 0;
      }
    /*!
       Same as above for constant objects
      */
    const T *operator-> () const
      {
      if (HandlePtr) return HandlePtr->data;
      _handle_error ("HANDLE: invalid operation");
// kludge: Some compilers want a return value
      return 0;
      }
    /*!
       Returns a reference to the handle's object.
       If the handle points to 0, program execution is stopped.
      */
    T &operator* ()
      {
      if (HandlePtr) return (*(HandlePtr->data));
      _handle_error ("HANDLE: invalid operation");
// kludge: Some compilers want a return value
      return (*(HandlePtr->data));
      }
    /**
       Same as above for constant objects
      */
    const T &operator* () const
      {
      if (HandlePtr) return (*(HandlePtr->data));
      _handle_error ("HANDLE: invalid operation");
// kludge: Some compilers want a return value
      return (*(HandlePtr->data));
      }

    /*!
       Returns false, if the handle points to 0, else true.
      */
    bool Valid () const
      { return (HandlePtr ? true : false); }
    /*!
       Returns false, if the handle points to 0, else true.
      */
    operator bool () const
      { return (HandlePtr ? true: false); }

    T* Pointer() const
      {
      if (HandlePtr) return (HandlePtr->data);
      return 0;
      }
  };


/**
   \class TMPHANDLE kernel/handle.h kernel/handle.h
   Reference-like handle class.
*/
template <class T> class TMPHANDLE
  {
  private:
    T *Ptr;

  public:
    /*! */
    TMPHANDLE () 
      : Ptr (0) {}
    /*! */
    TMPHANDLE (const TMPHANDLE &orig)
      : Ptr (orig.Ptr) {}
    /*! */
    TMPHANDLE (const HANDLE<T> orig)
      : Ptr (orig.Pointer()) {}
    /*! */
    ~TMPHANDLE ()
      {}
    /*! */
    TMPHANDLE &operator= (const HANDLE<T> orig)
      { Ptr = orig.Pointer(); return *this; }    
    /*! */
    TMPHANDLE &operator= (const TMPHANDLE orig)
      { Ptr = orig.Ptr; return *this; }    
    /*! */
    TMPHANDLE &operator= (T *contents)
      { Ptr = contents; return *this; }
    /*! */
    T *operator-> ()
      { return Ptr; }
    /*! */
    const T *operator-> () const
      { return Ptr; }
    /*! */
    T &operator* ()
      { return (*Ptr); }
    /*! */
    const T &operator* () const
      { return (*Ptr); }
    /*! */
    bool Valid () const
      { return (Ptr ? true : false); }
    /*! */
    operator bool () const
      { return (Ptr ? true: false); }
  };

//RAYPP_NAMESPACE_END (RAYPP)

#endif
